from langchain_ai21.chat.chat_adapter import (
    ChatAdapter,
    JambaChatCompletionsAdapter,
)


def create_chat_adapter(model: str) -> ChatAdapter:
    """Create a chat adapter based on the model.

    Args:
        model: The model to create the chat adapter for.

    Returns:
        The chat adapter.
    """
    if "jamba" in model:
        return JambaChatCompletionsAdapter()

    raise ValueError(f"Model {model} not supported.")



from typing import Any, Dict, List, Literal, Optional, Type

from ai21.models.maestro.run import Budget, RunResponse, ToolType
from langchain_core.callbacks import (
    AsyncCallbackManagerForLLMRun,
    CallbackManagerForLLMRun,
)
from langchain_core.language_models.chat_models import BaseChatModel
from langchain_core.messages import AIMessage, BaseMessage
from langchain_core.outputs import ChatGeneration, ChatResult
from typing_extensions import TypedDict

from langchain_ai21.ai21_base import AI21Base


class FileSearchToolResource(TypedDict, total=False):
    retrieval_similarity_threshold: Optional[float]
    labels: Optional[List[str]]
    labels_filter_mode: Optional[Literal["AND", "OR"]]
    labels_filter: Optional[dict]
    file_ids: Optional[List[str]]
    retrieval_strategy: Optional[str]
    max_neighbors: Optional[int]


class WebSearchToolResource(TypedDict, total=False):
    urls: Optional[List[str]]


class ToolResources(TypedDict, total=False):
    file_search: Optional[FileSearchToolResource]
    web_search: Optional[WebSearchToolResource]


class ChatMaestro(BaseChatModel, AI21Base):
    """Chat model using Maestro LLM."""

    output_type: Optional[Dict[str, Any]] = None
    """Optional dictionary specifying the output type."""

    models: Optional[List[str]] = None
    """Optional list of model names to use.
        Available models https://github.com/AI21Labs/ai21-python?tab=readme-ov-file#model-types"""

    tools: Optional[List[Dict[str, ToolType]]] = None
    """Optional list of tools."""

    tool_resources: Optional[ToolResources] = None
    """Optional resources for the tools."""

    context: Optional[Dict[str, Any]] = None
    """Optional dictionary providing context for the chat."""

    budget: Optional[Budget] = None
    """Optional budget constraints for the chat."""

    poll_interval_sec: Optional[float] = 1
    """Interval in seconds for polling the run status."""

    poll_timeout_sec: Optional[float] = 120
    """Timeout in seconds for polling the run status."""

    @property
    def _llm_type(self) -> str:
        """Return the type of LLM."""
        return "chat-maestro"

    def _call(self, messages: List[BaseMessage], **kwargs: Any) -> RunResponse:
        """API call to Maestro."""
        payload = self._prepare_payload(messages, **kwargs)
        result = self.client.beta.maestro.runs.create_and_poll(**payload)
        if result.status != "completed":
            raise RuntimeError(f"Maestro run failed with status: {result.status}")

        return result

    async def _acall(self, messages: List[BaseMessage], **kwargs: Any) -> RunResponse:
        """Asynchronous API call to Maestro."""
        payload = self._prepare_payload(messages, **kwargs)
        result = await self.async_client.beta.maestro.runs.create_and_poll(**payload)
        if result.status != "completed":
            raise RuntimeError(f"Maestro run failed with status: {result.status}")

        return result

    def _generate(
        self,
        messages: list[BaseMessage],
        stop: Optional[list[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> ChatResult:
        """Generates a response using Maestro LLM."""
        response_data = self._call(messages, **kwargs)
        return self._handle_chat_result(response_data)

    async def _agenerate(
        self,
        messages: list[BaseMessage],
        stop: Optional[list[str]] = None,
        run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> ChatResult:
        """Asynchronous agent call to Maestro."""
        response_data = await self._acall(messages, **kwargs)
        return self._handle_chat_result(response_data)

    @staticmethod
    def _prepare_payload(messages: List[BaseMessage], **kwargs: Any) -> Dict[str, Any]:
        """Prepare the payload for the API call with validation."""
        formatted_messages = [
            {"role": "user", "content": message.content} for message in messages
        ]
        payload = {"input": formatted_messages, **kwargs}

        requirements = payload.pop("requirements", [])
        if requirements:
            ChatMaestro.validate_list(requirements, "requirements")
            payload["requirements"] = [
                {"name": req, "description": req} for req in requirements
            ]

        variables = payload.pop("variables", [])
        if variables:
            ChatMaestro.validate_list(variables, "variables")
            variables_str = " ".join(variables)
            payload["requirements"] = payload.get("requirements", []) + [
                {
                    "name": "output should contain only these variables:"
                    f" {variables_str}",
                    "description": variables_str,
                }
            ]

        return payload

    @staticmethod
    def validate_list(obj: List[str], obj_name: str, expected_type: Type = str) -> None:
        """Validate that obj is a list of the expected type."""
        if obj is not None and (
            not isinstance(obj, list)
            or any(not isinstance(var, expected_type) for var in obj)
        ):
            raise ValueError(f"{obj_name} must be a list of {expected_type.__name__}")

    @staticmethod
    def _handle_chat_result(response_data: RunResponse) -> ChatResult:
        """Handle the response data from the Maestro run."""
        ai_message = AIMessage(content=response_data.result)
        generation = ChatGeneration(message=ai_message)
        return ChatResult(generations=[generation])



import asyncio
from functools import partial
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Type,
    Union,
)

from langchain_core.callbacks import (
    AsyncCallbackManagerForLLMRun,
    CallbackManagerForLLMRun,
)
from langchain_core.language_models import LanguageModelInput
from langchain_core.language_models.chat_models import (
    BaseChatModel,
    LangSmithParams,
    generate_from_stream,
)
from langchain_core.messages import BaseMessage
from langchain_core.outputs import ChatGeneration, ChatGenerationChunk, ChatResult
from langchain_core.runnables import Runnable
from langchain_core.tools import BaseTool
from langchain_core.utils.function_calling import convert_to_openai_tool
from pydantic import ConfigDict, model_validator
from typing_extensions import Self

from langchain_ai21.ai21_base import AI21Base
from langchain_ai21.chat.chat_adapter import ChatAdapter
from langchain_ai21.chat.chat_factory import create_chat_adapter


class ChatAI21(BaseChatModel, AI21Base):
    """ChatAI21 chat model. Different model types support different parameters and
    different parameter values. Please read the [AI21 reference documentation]
    (https://docs.ai21.com/reference) for your model to understand which parameters
    are available.

    Example:
        .. code-block:: python

            from langchain_ai21 import ChatAI21


            model = ChatAI21(
                # defaults to os.environ.get("AI21_API_KEY")
                api_key="my_api_key"
            )
    """

    model: str
    """Model type you wish to interact with. 
        You can view the options at https://github.com/AI21Labs/ai21-python?tab=readme-ov-file#model-types"""

    stop: Optional[List[str]] = None
    """Default stop sequences."""

    max_tokens: int = 512
    """The maximum number of tokens to generate for each response."""

    temperature: float = 0.4
    """A value controlling the "creativity" of the model's responses."""

    top_p: float = 1
    """A value controlling the diversity of the model's responses."""

    n: int = 1
    """Number of chat completions to generate for each prompt."""
    streaming: bool = False

    _chat_adapter: ChatAdapter

    @model_validator(mode="after")
    def validate_environment(self) -> Self:
        """Validate the environment."""
        model = self.model
        self._chat_adapter = create_chat_adapter(model)
        return self

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
    )

    @property
    def _llm_type(self) -> str:
        """Return type of chat model."""
        return "chat-ai21"

    @property
    def _default_params(self) -> Mapping[str, Any]:
        base_params = {
            "model": self.model,
            "max_tokens": self.max_tokens,
            "temperature": self.temperature,
            "top_p": self.top_p,
            "n": self.n,
        }
        if self.stop:
            base_params["stop_sequences"] = self.stop

        return base_params

    def _get_ls_params(
        self, stop: Optional[List[str]] = None, **kwargs: Any
    ) -> LangSmithParams:
        """Get standard params for tracing."""
        params = self._get_invocation_params(stop=stop, **kwargs)
        ls_params = LangSmithParams(
            ls_provider="ai21",
            ls_model_name=self.model,
            ls_model_type="chat",
            ls_temperature=params.get("temperature", self.temperature),
        )
        if ls_max_tokens := params.get("max_tokens", self.max_tokens):
            ls_params["ls_max_tokens"] = ls_max_tokens
        if ls_stop := stop or params.get("stop", None) or self.stop:
            ls_params["ls_stop"] = ls_stop
        return ls_params

    def _build_params_for_request(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> Mapping[str, Any]:
        params = {}
        converted_messages = self._chat_adapter.convert_messages(messages)

        if stop is not None:
            if "stop" in kwargs:
                raise ValueError("stop is defined in both stop and kwargs")
            params["stop_sequences"] = stop

        return {
            **converted_messages,
            **self._default_params,
            **params,
            **kwargs,
        }

    def _generate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        stream: Optional[bool] = None,
        **kwargs: Any,
    ) -> ChatResult:
        should_stream = stream or self.streaming

        if should_stream:
            return self._handle_stream_from_generate(
                messages=messages,
                stop=stop,
                run_manager=run_manager,
                **kwargs,
            )

        params = self._build_params_for_request(
            messages=messages,
            stop=stop,
            stream=should_stream,
            **kwargs,
        )

        messages = self._chat_adapter.call(self.client, **params)
        generations = [ChatGeneration(message=message) for message in messages]

        return ChatResult(generations=generations)

    def _handle_stream_from_generate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> ChatResult:
        stream_iter = self._stream(
            messages=messages,
            stop=stop,
            run_manager=run_manager,
            **kwargs,
        )
        return generate_from_stream(stream_iter)

    def _stream(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> Iterator[ChatGenerationChunk]:
        params = self._build_params_for_request(
            messages=messages,
            stop=stop,
            stream=True,
            **kwargs,
        )

        for chunk in self._chat_adapter.call(self.client, **params):
            if run_manager and isinstance(chunk.message.content, str):
                run_manager.on_llm_new_token(token=chunk.message.content, chunk=chunk)
            yield chunk

    async def _agenerate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> ChatResult:
        return await asyncio.get_running_loop().run_in_executor(
            None, partial(self._generate, **kwargs), messages, stop, run_manager
        )

    def _get_system_message_from_message(self, message: BaseMessage) -> str:
        if not isinstance(message.content, str):
            raise ValueError(
                f"System Message must be of type str. Got {type(message.content)}"
            )

        return message.content

    def bind_tools(
        self,
        tools: Sequence[Union[Dict[str, Any], Type, Callable, BaseTool]],
        **kwargs: Any,
    ) -> Runnable[LanguageModelInput, BaseMessage]:
        formatted_tools = [convert_to_openai_tool(tool) for tool in tools]
        return super().bind(tools=formatted_tools, **kwargs)



from typing import Any, Optional

from ai21 import AI21Client, AsyncAI21Client
from langchain_core.utils import from_env, secret_from_env
from pydantic import BaseModel, ConfigDict, Field, SecretStr, model_validator
from typing_extensions import Self

_DEFAULT_TIMEOUT_SEC = 300


class AI21Base(BaseModel):
    """Base class for AI21 models."""

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
    )

    client: Any = Field(default=None, exclude=True)  #: :meta private:

    async_client: Any = Field(default=None, exclude=True)
    """Asynchronous client for API calls."""

    api_key: SecretStr = Field(
        default_factory=secret_from_env("AI21_API_KEY", default="")
    )
    """API key for AI21 API."""
    api_host: str = Field(
        default_factory=from_env(
            "AI21_API_URL", default="https://api.ai21.com/studio/v1"
        )
    )
    """Host URL"""
    timeout_sec: float = Field(
        default_factory=lambda: float(
            from_env("AI21_TIMEOUT_SEC", default=str(_DEFAULT_TIMEOUT_SEC))()
        )
    )
    """Timeout in seconds.
    
    If not set, it will default to the value of the environment 
    variable `AI21_TIMEOUT_SEC` or 300 seconds.
    """
    num_retries: Optional[int] = None
    """Maximum number of retries for API requests before giving up."""

    @model_validator(mode="after")
    def post_init(self) -> Self:
        api_key = self.api_key
        api_host = self.api_host
        timeout_sec = self.timeout_sec
        if (self.client or None) is None:
            self.client = AI21Client(
                api_key=api_key.get_secret_value(),
                api_host=api_host,
                timeout_sec=None if timeout_sec is None else float(timeout_sec),
                via="langchain",
            )

        return self

    @model_validator(mode="after")
    def init_async_client(self) -> Self:
        api_key = self.api_key
        api_host = self.api_host
        timeout_sec = self.timeout_sec
        if (self.async_client or None) is None:
            self.async_client = AsyncAI21Client(
                api_key=api_key.get_secret_value(),
                api_host=api_host,
                timeout_sec=None if timeout_sec is None else float(timeout_sec),
                via="langchain",
            )

        return self



from __future__ import annotations

import json
from abc import ABC, abstractmethod
from typing import Any, Dict, Iterator, List, Literal, Optional, Union, overload

from ai21.models import RoleType
from ai21.models.chat import AssistantMessage as AI21AssistantMessage
from ai21.models.chat import ChatCompletionChunk, ChatMessageParam
from ai21.models.chat import ChatMessage as AI21ChatMessage
from ai21.models.chat import SystemMessage as AI21SystemMessage
from ai21.models.chat import ToolCall as AI21ToolCall
from ai21.models.chat import ToolFunction as AI21ToolFunction
from ai21.models.chat import ToolMessage as AI21ToolMessage
from ai21.models.chat import UserMessage as AI21UserMessage
from ai21.stream.stream import Stream as AI21Stream
from langchain_core.messages import (
    AIMessage,
    AIMessageChunk,
    BaseMessage,
    BaseMessageChunk,
    HumanMessage,
    SystemMessage,
    ToolCall,
    ToolMessage,
)
from langchain_core.messages.ai import UsageMetadata
from langchain_core.output_parsers.openai_tools import parse_tool_call
from langchain_core.outputs import ChatGenerationChunk

_ChatMessageTypes = Union[AI21ChatMessage]
_SYSTEM_ERR_MESSAGE = "System message must be at beginning of message list."
_ROLE_TYPE = Union[str, RoleType]


class ChatAdapter(ABC):
    """Common interface for the different Chat models available in AI21.

    It converts LangChain messages to AI21 messages.
    Calls the appropriate AI21 model API with the converted messages.
    """

    @abstractmethod
    def convert_messages(
        self,
        messages: List[BaseMessage],
    ) -> Dict[str, Any]:
        pass

    def _convert_message_to_ai21_message(
        self,
        message: BaseMessage,
    ) -> _ChatMessageTypes:
        role = self._parse_role(message)
        return self._chat_message(role=role, message=message)

    def _parse_role(self, message: BaseMessage) -> _ROLE_TYPE:
        if isinstance(message, SystemMessage):
            return RoleType.SYSTEM

        if isinstance(message, HumanMessage):
            return RoleType.USER

        if isinstance(message, AIMessage):
            return RoleType.ASSISTANT

        if isinstance(message, ToolMessage):
            return RoleType.TOOL

        # if it gets here, we rely on the server to handle the role type
        return message.type

    @abstractmethod
    def _chat_message(
        self,
        role: _ROLE_TYPE,
        message: BaseMessage,
    ) -> _ChatMessageTypes:
        pass

    @overload
    def call(
        self,
        client: Any,
        stream: Literal[True],
        **params: Any,
    ) -> Iterator[ChatGenerationChunk]:
        pass

    @overload
    def call(
        self,
        client: Any,
        stream: Literal[False],
        **params: Any,
    ) -> List[BaseMessage]:
        pass

    @abstractmethod
    def call(
        self,
        client: Any,
        stream: Literal[True] | Literal[False],
        **params: Any,
    ) -> List[BaseMessage] | Iterator[ChatGenerationChunk]:
        pass

    def _get_system_message_from_message(self, message: BaseMessage) -> str:
        if not isinstance(message.content, str):
            raise ValueError(
                f"System Message must be of type str. Got {type(message.content)}"
            )

        return message.content


class JambaChatCompletionsAdapter(ChatAdapter):
    """Adapter for Jamba Chat Completions."""

    def convert_messages(self, messages: List[BaseMessage]) -> Dict[str, Any]:
        return {
            "messages": [
                self._convert_message_to_ai21_message(message) for message in messages
            ],
        }

    def _convert_lc_tool_calls_to_ai21_tool_calls(
        self, tool_calls: List[ToolCall]
    ) -> Optional[List[AI21ToolCall]]:
        """
        Convert Langchain ToolCalls to AI21 ToolCalls.
        """
        ai21_tool_calls: List[AI21ToolCall] = []
        for lc_tool_call in tool_calls:
            if "id" not in lc_tool_call or not lc_tool_call["id"]:
                raise ValueError("Tool call ID is missing or empty.")

            ai21_tool_call = AI21ToolCall(
                id=lc_tool_call["id"],
                type="function",
                function=AI21ToolFunction(
                    name=lc_tool_call["name"],
                    arguments=json.dumps(lc_tool_call["args"]),
                ),
            )
            ai21_tool_calls.append(ai21_tool_call)

        return ai21_tool_calls

    def _get_content_as_string(self, base_message: BaseMessage) -> str:
        if isinstance(base_message.content, str):
            return base_message.content
        elif isinstance(base_message.content, list):
            return "\n".join(str(item) for item in base_message.content)
        else:
            raise ValueError("Unsupported content type")

    def _chat_message(
        self,
        role: _ROLE_TYPE,
        message: BaseMessage,
    ) -> ChatMessageParam:
        content = self._get_content_as_string(message)

        if isinstance(message, AIMessage):
            return AI21AssistantMessage(
                tool_calls=self._convert_lc_tool_calls_to_ai21_tool_calls(
                    message.tool_calls
                ),
                content=content or None,
            )
        if isinstance(message, ToolMessage):
            return AI21ToolMessage(
                tool_call_id=message.tool_call_id,
                content=content,
            )
        if isinstance(message, HumanMessage):
            return AI21UserMessage(
                content=content,
            )
        if isinstance(message, SystemMessage):
            return AI21SystemMessage(
                content=content,
            )
        return AI21ChatMessage(
            role=role.value if isinstance(role, RoleType) else role,
            content=content,
        )

    @overload
    def call(
        self,
        client: Any,
        stream: Literal[True],
        **params: Any,
    ) -> Iterator[ChatGenerationChunk]: ...

    @overload
    def call(
        self,
        client: Any,
        stream: Literal[False],
        **params: Any,
    ) -> List[BaseMessage]: ...

    def call(
        self,
        client: Any,
        stream: Literal[True] | Literal[False],
        **params: Any,
    ) -> List[BaseMessage] | Iterator[ChatGenerationChunk]:
        response = client.chat.completions.create(stream=stream, **params)

        if stream:
            return self._stream_response(response)

        ai_messages: List[BaseMessage] = []
        for message in response.choices:
            if message.message.tool_calls:
                tool_calls = [
                    parse_tool_call(tool_call.model_dump(), return_id=True)
                    for tool_call in message.message.tool_calls
                ]
                ai_messages.append(AIMessage("", tool_calls=tool_calls))
            else:
                ai_messages.append(AIMessage(message.message.content))

        return ai_messages

    def _stream_response(
        self,
        response: AI21Stream[ChatCompletionChunk],
    ) -> Iterator[ChatGenerationChunk]:
        for chunk in response:
            converted_message = self._convert_ai21_chunk_to_chunk(chunk)
            yield ChatGenerationChunk(message=converted_message)

    def _convert_ai21_chunk_to_chunk(
        self,
        chunk: ChatCompletionChunk,
    ) -> BaseMessageChunk:
        usage = chunk.usage
        content = chunk.choices[0].delta.content or ""

        if usage is None:
            return AIMessageChunk(
                content=content,
            )

        return AIMessageChunk(
            content=content,
            usage_metadata=UsageMetadata(
                input_tokens=usage.prompt_tokens,
                output_tokens=usage.completion_tokens,
                total_tokens=usage.total_tokens,
            ),
        )



import os
import uuid
from typing import Annotated, Any, List

from langchain_core.messages import (
    AIMessage,
    BaseMessage,
    HumanMessage,
    SystemMessage,
    ToolMessage,
)
from langchain_core.runnables import RunnableConfig
from langchain_openai import ChatOpenAI
from langgraph.checkpoint.memory import MemorySaver
from langgraph.graph import END, START, StateGraph
from langgraph.graph.message import add_messages
from pydantic import BaseModel
from typing_extensions import TypedDict

from langchain_ai21.chat.chat_maestro import ChatMaestro

os.environ["OPENAI_API_KEY"] = "openai_api_key"
os.environ["AI21_API_KEY"] = "ai21_api_key"
os.environ["AI21_API_HOST"] = "ai21_api_host"

template = """Your job is to get information from a user about what type of email
 template they want to create.

You should get the following information from them:

- What the objective of the email is
- What variables will be passed into the email
- Any requirements that the output MUST adhere to

If you are not able to discern this info, ask them to clarify! Do not attempt to wildly
 guess.

After you are able to discern all the information, call the relevant tool."""


def get_messages_info(messages: list[SystemMessage]) -> List[SystemMessage]:
    return [SystemMessage(content=template)] + messages


class EmailInstructions(BaseModel):
    """Instructions on how to generate the email template."""

    objective: str
    variables: List[str]
    requirements: List[str]


llm_prompt = ChatMaestro()
llm_info = ChatOpenAI(temperature=0)
llm_with_tool = llm_info.bind_tools([EmailInstructions])


def info_chain(state: dict) -> dict:
    messages = get_messages_info(state["messages"])
    response = llm_with_tool.invoke(messages)
    return {"messages": [response]}


# Function to get the messages for the prompt
# Will only get messages AFTER the tool call
def get_prompt_messages_maestro(
    messages: list[BaseMessage],
) -> tuple[list[BaseMessage], dict[str, Any]]:
    tool_call = {}
    other_msgs = []
    for m in messages:
        if isinstance(m, AIMessage) and m.tool_calls:
            tool_call = m.tool_calls[0]["args"]
        elif isinstance(m, ToolMessage):
            continue
        elif tool_call is not None:
            other_msgs.append(m)
    return other_msgs, tool_call


def prompt_gen_chain(state: dict) -> dict:
    messages, tool_call = get_prompt_messages_maestro(state["messages"])
    objective = tool_call.pop("objective")
    variables = tool_call.get("variables")
    maestro_input = f"generate a prompt that meets the following objective: {objective}"

    if variables:
        maestro_input += f" with the following variables: {variables}"

    response = llm_prompt.invoke([SystemMessage(maestro_input)], **tool_call)
    return {"messages": [response]}


def get_state(state: dict) -> str:
    messages = state["messages"]
    if isinstance(messages[-1], AIMessage) and messages[-1].tool_calls:
        return "add_tool_message"
    elif not isinstance(messages[-1], HumanMessage):
        return END
    return "info"


class State(TypedDict):
    messages: Annotated[list, add_messages]


memory = MemorySaver()
workflow = StateGraph(State)
workflow.add_node("info", info_chain)
workflow.add_node("prompt", prompt_gen_chain)


@workflow.add_node
def add_tool_message(state: State) -> dict:
    return {
        "messages": [
            ToolMessage(
                content="Generating email template...",
                tool_call_id=state["messages"][-1].tool_calls[0]["id"],
            )
        ]
    }


workflow.add_conditional_edges("info", get_state, ["add_tool_message", "info", END])
workflow.add_edge("add_tool_message", "prompt")
workflow.add_edge("prompt", END)
workflow.add_edge(START, "info")
graph = workflow.compile(checkpointer=memory)

cached_human_responses = [
    "hi!",
    "write a prompt",
    "1 email to reach out to sell maestro the next ai tool, 2 recipient_first_name,"
    " recipient_last_name, 3 the email should contain 5 rows",
    "Sritala Hollinger",
    "q",
]

cached_response_index = 0
config: RunnableConfig = {
    "configurable": {
        "thread_id": str(uuid.uuid4()),
    }
}

while True:
    try:
        user = input("User (q/Q to quit): ")
    except EOFError:
        user = cached_human_responses[cached_response_index]
        cached_response_index += 1
    print(f"User (q/Q to quit): {user}")
    if user in {"q", "Q"}:
        print("AI: Bye bye")
        break
    output = None
    for output in graph.stream(
        {"messages": [HumanMessage(content=user)]}, config=config, stream_mode="updates"
    ):
        last_message = next(iter(output.values()))["messages"][-1]
        last_message.pretty_print()

    if output and "prompt" in output:
        print("Done!")
